@@ -51,25 +51,17 @@
    GIF        |  .gif            |     X     |     
    BMP        |  .gif            |     X     |     
 */
-#include "inr.h"
-#include "gif.h"
-#include "gis.h"
-#include "pnm.h"
-#include "bmp.h"
-#include "iris.h"
-#include "analyze.h"
+#include <CGAL/ImageIO/inr.h>
+#include <CGAL/ImageIO/gif.h>
+#include <CGAL/ImageIO/gis.h>
+#include <CGAL/ImageIO/pnm.h>
+#include <CGAL/ImageIO/bmp.h>
+#include <CGAL/ImageIO/iris.h>
+#include <CGAL/ImageIO/analyze.h>
 #ifdef MINC_FILES
-#include "mincio.h"
+#include <CGAL/ImageIO/mincio.h>
 #endif
 
-
-
-/** the first file format is initialized to null */
-static PTRIMAGE_FORMAT firstFormat=NULL;
-
-/** the Inrimage file format (default format) is initialized to null */
-static PTRIMAGE_FORMAT InrimageFormat=NULL;
-
 struct Remove_supported_file_format {
   ~Remove_supported_file_format()
   {
@@ -77,9 +69,52 @@ struct Remove_supported_file_format {
   }
 };
 
+#ifdef CGAL_HEADER_ONLY
+
+inline PTRIMAGE_FORMAT & get_static_firstFormat()
+{
+  static PTRIMAGE_FORMAT firstFormat = NULL;
+  return firstFormat;
+}
+
+inline PTRIMAGE_FORMAT & get_static_inrimageFormat()
+{
+  static PTRIMAGE_FORMAT inrimageFormat = NULL;
+  return inrimageFormat;
+}
+
+inline Remove_supported_file_format & get_static_rsff()
+{
+  static Remove_supported_file_format rsff;
+  return rsff;
+}
+// Dummy call to get_static_rsff(), otherwise it would not get instanced
+CGAL_UNUSED static Remove_supported_file_format &rsff_dummy_ref = get_static_rsff();
+
+
+#else // not header-only
+
+/** the first file format is initialized to null */
+static PTRIMAGE_FORMAT firstFormat = NULL;
+inline PTRIMAGE_FORMAT & get_static_firstFormat()
+{
+  return firstFormat;
+}
+
+/** the Inrimage file format (default format) is initialized to null */
+static PTRIMAGE_FORMAT InrimageFormat = NULL;
+inline PTRIMAGE_FORMAT & get_static_inrimageFormat()
+{
+  return InrimageFormat;
+}
+
 static Remove_supported_file_format rsff;
+inline Remove_supported_file_format & get_static_rsff()
+{
+  return rsff;
+}
 
-
+#endif
 
 /*--------------------------------------------------
  *
@@ -110,6 +145,11 @@ void ImageIO_free(void *m) {
 /* call deallocation routine */
 
 
+CGAL_INLINE_FUNCTION
+unsigned int ImageIO_limit_len(size_t to_be_read)
+{
+  return (unsigned int)(std::min)(to_be_read, size_t(1u<<30));
+}
 
 /* mimics fwrite() function.
    According to _openWriteImage(), openMode will has one 
@@ -120,6 +160,9 @@ void ImageIO_free(void *m) {
 */
 CGAL_INLINE_FUNCTION
 size_t ImageIO_write(const _image *im, const void *buf, size_t len) {
+  size_t to_be_written = len;
+  int l = -1;
+  char *b = (char*)buf;
 
   switch(im->openMode) {
   default :
@@ -127,20 +170,41 @@ size_t ImageIO_write(const _image *im, const void *buf, size_t len) {
     return 0;
   case OM_STD :
 #ifdef CGAL_USE_ZLIB
-    return gzwrite(im->fd, (void *) buf, len);
+    while ( (to_be_written > 0) && ((l = gzwrite(im->fd, (void *) b, ImageIO_limit_len(to_be_written))) > 0) ) {
+      to_be_written -= l;
+      b += l;
+    }
 #else 
-    return fwrite(buf, 1, len, im->fd);
+    while ( (to_be_written > 0) && ((l = fwrite( b, 1, ImageIO_limit_len(to_be_written), im->fd )) > 0) ) {
+      to_be_written -= l;
+      b += l;
+    }
 #endif
+    return ( len - to_be_written );
 #ifdef CGAL_USE_ZLIB
   case OM_GZ :
-    return gzwrite(im->fd, (void *) buf, len);
+    while ( (to_be_written > 0) && ((l = gzwrite(im->fd, (void *) b, ImageIO_limit_len(to_be_written))) > 0) ) {
+      to_be_written -= l;
+      b += l;
+    }
+    if(l<0)
+    {
+      int errnum;
+      fprintf(stderr, "zlib error: %s\n", gzerror(im->fd, &errnum));
+    }
+    return ( len - to_be_written );
+#else
+  case OM_FILE :
+    while ( (to_be_written > 0) && ((l = fwrite( b, 1, ImageIO_limit_len(to_be_written), im->fd )) > 0) ) {
+      to_be_written -= l;
+      b += l;
+    }
+    return ( len - to_be_written );
 #endif
-  case OM_FILE:
-    return fwrite(buf, 1, len, (FILE*)im->fd);
   }
-  //return 0;
 }
 
+CGAL_INLINE_FUNCTION
 size_t ImageIO_limit_read(size_t to_be_read)
 {
   return (std::min)(to_be_read, size_t(1u<<30));
@@ -165,12 +229,12 @@ size_t ImageIO_read(const _image *im, void *buf, size_t len)
     return 0;
   case OM_STD :
 #ifdef CGAL_USE_ZLIB
-    while ( (to_be_read > 0) && ((l = gzread(im->fd, (void *) b, ImageIO_limit_read(to_be_read))) > 0) ) {
+    while ( (to_be_read > 0) && ((l = gzread(im->fd, (void *) b, ImageIO_limit_len(to_be_read))) > 0) ) {
       to_be_read -= l;
       b += l;
     }
 #else 
-    while ( (to_be_read > 0) && ((l = fread( b, 1, ImageIO_limit_read(to_be_read), im->fd )) > 0) ) {
+    while ( (to_be_read > 0) && ((l = fread( b, 1, ImageIO_limit_len(to_be_read), im->fd )) > 0) ) {
       to_be_read -= l;
       b += l;
     }
@@ -178,7 +242,7 @@ size_t ImageIO_read(const _image *im, void *buf, size_t len)
     return ( len - to_be_read );
 #ifdef CGAL_USE_ZLIB
   case OM_GZ :
-    while ( (to_be_read > 0) && ((l = gzread(im->fd, (void *) b, ImageIO_limit_read(to_be_read))) > 0) ) {
+    while ( (to_be_read > 0) && ((l = gzread(im->fd, (void *) b, ImageIO_limit_len(to_be_read))) > 0) ) {
       to_be_read -= l;
       b += l;
     }
@@ -190,7 +254,7 @@ size_t ImageIO_read(const _image *im, void *buf, size_t len)
     return ( len - to_be_read );
 #else
   case OM_FILE :
-    while ( (to_be_read > 0) && ((l = fread( b, 1, ImageIO_limit_read(to_be_read), im->fd )) > 0) ) {
+    while ( (to_be_read > 0) && ((l = fread( b, 1, ImageIO_limit_len(to_be_read), im->fd )) > 0) ) {
       to_be_read -= l;
       b += l;
     }
@@ -240,7 +304,7 @@ char *ImageIO_gets( const _image *im, char *str, int size )
 
 
 CGAL_INLINE_FUNCTION
-int ImageIO_seek( const _image *im, long offset, int whence ) {
+long ImageIO_seek( const _image *im, long offset, int whence ) {
   switch(im->openMode) {
   case OM_CLOSE :
   default :
@@ -511,7 +575,7 @@ _image *_initImage() {
   im->imageFormat = NULL;
 
   /** eventually initializes the supported file formats */
-  if (firstFormat==NULL)
+  if (get_static_firstFormat()==NULL)
     initSupportedFileFormat();
   /* return image descriptor */
   return im;
@@ -569,7 +633,7 @@ _image *_createImage(int x, int y, int z, int v,
   im->imageFormat = NULL;
 
   /** eventually initializes the supported file formats */
-  if (firstFormat==NULL)
+  if (get_static_firstFormat()==NULL)
     initSupportedFileFormat();
   /* return image descriptor */
   return im;
@@ -583,9 +647,9 @@ void _get_image_bounding_box(_image* im,
   *x_min = im->tx;
   *y_min = im->ty;
   *z_min = im->tz;
-  *x_max = (im->xdim - 1.0f)*(im->vx) + *x_min ;
-  *y_max = (im->ydim - 1.0f)*(im->vy) + *y_min ;
-  *z_max = (im->zdim - 1.0f)*(im->vz) + *z_min ;
+  *x_max = (double(im->xdim) - 1.0)*(im->vx) + *x_min ;
+  *y_max = (double(im->ydim) - 1.0)*(im->vy) + *y_min ;
+  *z_max = (double(im->zdim) - 1.0)*(im->vz) + *z_min ;
 }
 
 /* Free an image descriptor */
@@ -790,7 +854,7 @@ CGAL_INLINE_FUNCTION
 int _writeImage(_image *im, const char *name_to_be_written ) {
 
   int r = ImageIO_NO_ERROR;
-  int length = 0;
+  std::size_t length = 0;
   char *name = NULL;
   char *baseName = NULL;
 
@@ -812,9 +876,9 @@ int _writeImage(_image *im, const char *name_to_be_written ) {
   /* what is the wanted format
    */
   if ( name == NULL ) {
-    im->imageFormat = InrimageFormat;
+    im->imageFormat = get_static_inrimageFormat();
   } else {
-    int i,extLength;
+    std::size_t i,extLength;
     PTRIMAGE_FORMAT f;
     char ext[IMAGE_FORMAT_NAME_LENGTH];
     char *ptr;
@@ -824,7 +888,7 @@ int _writeImage(_image *im, const char *name_to_be_written ) {
     im->imageFormat=NULL;
     length=strlen(name);
 
-    for(f=firstFormat;(f!=NULL)&& (im->imageFormat==NULL);f=f->next) {
+    for(f=get_static_firstFormat();(f!=NULL)&& (im->imageFormat==NULL);f=f->next) {
       /* scan all extensions for that format */
       ptr=&f->fileExtension[0];
 
@@ -857,7 +921,7 @@ int _writeImage(_image *im, const char *name_to_be_written ) {
 
     if (!im->imageFormat) { 
       fprintf(stderr, "_writeImage: warning : unknown extension in %s = assuming Inrimage\n",name);
-      im->imageFormat=InrimageFormat;
+      im->imageFormat=get_static_inrimageFormat();
       baseName=strdup(name);
     }
   }
@@ -879,7 +943,7 @@ int _writeImage(_image *im, const char *name_to_be_written ) {
   if (im->imageFormat) {
 
     if (im->imageFormat->writeImage==NULL) {
-      im->imageFormat=InrimageFormat;
+      im->imageFormat=get_static_inrimageFormat();
     }
 
     if ( 0 ) {
@@ -984,7 +1048,7 @@ _image *_readImageHeaderAndGetError( const char *name_to_be_read, int *error )
      assume that stdin is inrimage
    */
   if(im->openMode == OM_STD) {
-    im->imageFormat=InrimageFormat;
+    im->imageFormat=get_static_inrimageFormat();
   }
   else {
    /* get magic string for disk files
@@ -993,7 +1057,7 @@ _image *_readImageHeaderAndGetError( const char *name_to_be_read, int *error )
     magic[4] = '\0';
     ImageIO_seek(im, 0L, SEEK_SET);
     /** test each format */
-    for(f=firstFormat;(f!=NULL)&& (im->imageFormat==NULL);f=f->next) {
+    for(f=get_static_firstFormat();(f!=NULL)&& (im->imageFormat==NULL);f=f->next) {
       /* test if it is the correct format based on magic and file extension */
       if (((*f->testImageFormat)(magic, name)) >=0) {
 	im->imageFormat=f;
@@ -1054,11 +1118,11 @@ static void _swapImageData( _image *im )
   unsigned char *ptr1, *ptr2, b[8];
   unsigned short int si, *ptr3, *ptr4;
   unsigned int        i, *ptr5, *ptr6;
-  int size, length;
+  std::size_t size, length;
   
   if( _getEndianness() != im->endianness) {
 
-    size = im->xdim * im->ydim * im->zdim * im->vdim * im->wdim;
+    size = std::size_t(im->xdim) * im->ydim * im->zdim * im->vdim * im->wdim;
     if ( size <= 0 ) return;
 
     length = size / im->wdim;
@@ -1077,7 +1141,7 @@ static void _swapImageData( _image *im )
       ptr3 = ptr4 = (unsigned short int *) im->data;
       while( length-- ) {
 	si = *ptr3++;
-	*ptr4++ = ((si >> 8) & 0xff) | (si << 8);
+	*ptr4++ = (unsigned short int)(((si >> 8) & 0xff) | (si << 8));
       }
     }
     
@@ -1377,10 +1441,10 @@ PTRIMAGE_FORMAT imageType(const char *fileName) {
   if(fileName) fclose( f );
 #endif
 
-  if (firstFormat==NULL)
+  if (get_static_firstFormat()==NULL)
     initSupportedFileFormat();
 
-  for(format=firstFormat;(format!=NULL);format=format->next) {
+  for(format=get_static_firstFormat();(format!=NULL);format=format->next) {
     /* test if it is the correct header based on magic and file extension */
     if (((*format->testImageFormat)(magic,fileName)) >=0) {
       return format;
@@ -1414,8 +1478,8 @@ int addImageFormat( PTRIMAGE_FORMAT format)
        (strlen(format->fileExtension)>0) &&
        (strlen(format->realName)>0) ) {
 
-    format->next=firstFormat;
-    firstFormat=format;
+    format->next=get_static_firstFormat();
+    get_static_firstFormat()=format;
     
     return 0;
 
@@ -1440,11 +1504,11 @@ int addImageFormatAtEnd( PTRIMAGE_FORMAT format)
 
     format->next = NULL;
     
-    if (firstFormat == NULL) {
-      firstFormat=format;
+    if (get_static_firstFormat() == NULL) {
+      get_static_firstFormat()=format;
     }
     else {
-      for(f=firstFormat;(f->next!=NULL);f=f->next)
+      for(f=get_static_firstFormat();(f->next!=NULL);f=f->next)
 	;
       f->next=format;    
     }
@@ -1465,7 +1529,7 @@ CGAL_INLINE_FUNCTION
 void initSupportedFileFormat() 
 {
   PTRIMAGE_FORMAT f;
-  if ( InrimageFormat == NULL ) {
+  if ( get_static_inrimageFormat() == NULL ) {
     f = createAnalyzeFormat();
     addImageFormatAtEnd( f );
     f = createBMPFormat();
@@ -1482,8 +1546,8 @@ void initSupportedFileFormat()
     addImageFormatAtEnd( f );
     f = createPpmFormat();
     addImageFormatAtEnd( f );
-    InrimageFormat = createInrimageFormat();
-    addImageFormat( InrimageFormat );
+    get_static_inrimageFormat() = createInrimageFormat();
+    addImageFormat( get_static_inrimageFormat() );
   }
 }
 
@@ -1491,7 +1555,7 @@ void initSupportedFileFormat()
 
 CGAL_INLINE_FUNCTION
 PTRIMAGE_FORMAT firstImageFormat() {
-  return firstFormat;
+  return get_static_firstFormat();
 }
 
 
@@ -1504,7 +1568,7 @@ void printSupportedFileFormat() {
 
   initSupportedFileFormat();
 
-  for(i=0, f=firstFormat;(f!=NULL);i++, f=f->next) {
+  for(i=0, f=get_static_firstFormat();(f!=NULL);i++, f=f->next) {
     if ( (f->testImageFormat) &&
 	 (f->readImageHeader) &&
 	 (strlen(f->fileExtension)>0) &&
@@ -1524,14 +1588,14 @@ void printSupportedFileFormat() {
 /** remove supported image formats */
 CGAL_INLINE_FUNCTION
 void removeSupportedFileFormat() {    
-  PTRIMAGE_FORMAT f=firstFormat;
+  PTRIMAGE_FORMAT f=get_static_firstFormat();
   
   while( f != NULL) {
     PTRIMAGE_FORMAT f_old = f;
     f = f->next;
     ImageIO_free( f_old);
   }
-  InrimageFormat=NULL;
+  get_static_inrimageFormat()=NULL;
 
 }
 
@@ -1545,10 +1609,10 @@ float triLinInterp(const _image* image,
                    float posz,
                    float value_outside /*= 0.f */)
 {
-  const int dimx = image->xdim;
-  const int dimy = image->ydim;
-  const int dimz = image->zdim;
-  const int dimxy = dimx*dimy;
+  const std::size_t dimx = image->xdim;
+  const std::size_t dimy = image->ydim;
+  const std::size_t dimz = image->zdim;
+  const std::size_t dimxy = dimx*dimy;
   
   if(posx < 0.f || posy < 0.f || posz < 0.f )
     return value_outside;
@@ -1569,10 +1633,10 @@ float triLinInterp(const _image* image,
   const int j2 = j1 + 1;
   const int k2 = k1 + 1;
 
-  const float KI2 = i2-posz;
-  const float KI1 = posz-i1;
-  const float KJ2 = j2-posy;
-  const float KJ1 = posy-j1;
+  const float KI2 = float(i2)-posz;
+  const float KI1 = posz-float(i1);
+  const float KJ2 = float(j2)-posy;
+  const float KJ1 = posy-float(j1);
 
   CGAL_IMAGE_IO_CASE
     (image,
@@ -1580,11 +1644,11 @@ float triLinInterp(const _image* image,
      return (((float)array[i1 * dimxy + j1 * dimx + k1] * KI2 +
               (float)array[i2 * dimxy + j1 * dimx + k1] * KI1) * KJ2 +
              ((float)array[i1 * dimxy + j2 * dimx + k1] * KI2 +
-              (float)array[i2 * dimxy + j2 * dimx + k1] * KI1) * KJ1) * (k2-posx)+
+              (float)array[i2 * dimxy + j2 * dimx + k1] * KI1) * KJ1) * (float(k2)-posx)+
      (((float)array[i1 * dimxy + j1 * dimx + k2] * KI2 +
        (float)array[i2 * dimxy + j1 * dimx + k2] * KI1) * KJ2 +
       ((float)array[i1 * dimxy + j2 * dimx + k2] * KI2 +
-       (float)array[i2 * dimxy + j2 * dimx + k2] * KI1) * KJ1) * (posx-k1);
+       (float)array[i2 * dimxy + j2 * dimx + k2] * KI1) * KJ1) * (posx-float(k1));
      );
   return 0.f;
 }
@@ -1610,9 +1674,9 @@ void convertImageTypeToFloat(_image* image){
   if(image->wordKind == WK_FLOAT && image->wdim == 4)
     return;
 
-  const unsigned int dimx = image->xdim;
-  const unsigned int dimy = image->ydim;
-  const unsigned int dimz = image->zdim;
+  const std::size_t dimx = image->xdim;
+  const std::size_t dimy = image->ydim;
+  const std::size_t dimz = image->zdim;
   
   float * array = (float*)ImageIO_alloc (dimx * dimy * dimz *sizeof(float));
   if (array == NULL ) {
