@@ -29,6 +29,7 @@
 
 #include <CGAL/Polygon_mesh_processing/internal/Side_of_triangle_mesh/Point_inside_vertical_ray_cast.h>
 
+#include <CGAL/Mesh_3/global_parameters.h>
 #include <CGAL/Mesh_3/Robust_intersection_traits_3.h>
 #include <CGAL/Mesh_3/Triangle_accessor_primitive.h>
 #include <CGAL/Triangle_accessor_3.h>
@@ -55,6 +56,10 @@
 # include <tbb/enumerable_thread_specific.h>
 #endif
 
+// To handle I/O for Surface_patch_index if that is a pair of `int` (the
+// default)
+#include <CGAL/internal/Mesh_3/Handle_IO_for_pair_of_int.h>
+
 namespace CGAL {
 
 namespace Mesh_3 {
@@ -286,9 +291,13 @@ public:
                    TriangleAccessor().triangles_end(bounding_polyhedron));
       tree_.build();
       bounding_tree_ =
+        bounding_polyhedron.empty() ?
+        0 :
         new AABB_tree_(TriangleAccessor().triangles_begin(bounding_polyhedron),
                        TriangleAccessor().triangles_end(bounding_polyhedron));
-      bounding_tree_->build();
+      if(!bounding_polyhedron.empty()) {
+        bounding_tree_->build();
+      }
     }
     else {
       tree_.rebuild(TriangleAccessor().triangles_begin(bounding_polyhedron),
@@ -366,6 +375,14 @@ public:
   }
 
 
+  /**
+   * Returns a bounding box of the domain
+   */
+  Bbox_3 bbox() const {
+    return tree_.bbox();
+  }
+
+
   /**
    * Returns true if point~\ccc{p} is in the domain. If \ccc{p} is in the
    *  domain, the parameter index is set to the index of the subdomain
@@ -473,7 +490,8 @@ public:
 #endif // not CGAL_MESH_3_NO_LONGER_CALLS_DO_INTERSECT_3
       {
 #ifndef CGAL_MESH_3_NO_LONGER_CALLS_DO_INTERSECT_3
-        CGAL_precondition(r_domain_.do_intersect_surface_object()(q));
+        CGAL_precondition(r_domain_.do_intersect_surface_object()(q)
+                          != boost::none);
 #endif // NOT CGAL_MESH_3_NO_LONGER_CALLS_DO_INTERSECT_3
 
         intersection = r_domain_.tree_.any_intersection(q);
